#ifndef ARITHPMD_H
#define ARITHPMD_H

#include "inc.h"
#include "nlohmann/json.hpp"
#include <opencv2/core/eigen.hpp>
#include <pcl/point_types.h>
#include <pcl/common/transforms.h>
#include "pcl/common/centroid.h" 
// #include <opencv2/aruco.hpp>

struct Screen {
    float width;    // Ļ(m)
    float height;  // Ļ߶(m)
    uint16_t cols;  // Ļֱ
    uint16_t rows;  // Ļֱ
};

struct ROI {
    uint32_t startRow;
    uint32_t startCol;
    uint32_t blockRows;
    uint32_t blockCols;
};

/*
 * Ҫ MatrixмҪҪʱ Array
 *  rowcolwidthheightͲxy
 * QtĹܾQtģĿ¼
 * ͼimgͼƬʾ
 * RתTLƽT任CVTVϵı任Cϵı任
 */

bool save_matrix_as_img(const MatrixXf& m, std::string path);

/*
 * ļ
 * add_index: Ƿ񽫵һлһбΪ±
 * orientation0һбΪ򣬵һбΪ
*/
bool save_matrix_as_txt(const MatrixXf& m, std::string path, bool add_index = false, int orientation = 0);

/*
 * 󱣴Ϊ ply ʽĵļ
 * params: , file path
 */
bool save_matrix_as_ply(const MatrixXf& m, std::string path);

/*
 * Ϊ
 */
MatrixXf matrix_to_home(const MatrixXf& origin);

/*
 * Rays from camera, for every pixel, a line cross this pixel and origin of camera coordinate
 * ÿ·Ϊһһƽ棬Ϊÿһÿ㶨壬һΪԭ㣬
 * ڶ(x,y,z)Ϊϵ¸zõά꣬Ծͷ
 * ڲΣֱʣZcamera_rays(ϵz=ֵ rays from camera)
 * camrea_rays ʽ: 3 x (img.h x img.w), ÿиʽΪ(x,y,z)ر˳Ϊң
 */
bool configCameraRay(const Matrix3f& cameraMatrix, const cv::Size& imageSize, float Z, MatrixXf& camera_rays);

/*
 * οƽĽ㣬οƽõ㷨ʽ
 * οƽ㣬οƽ淨ߣrays from camera潻
 * refPlane ʽ3 x (img.h x img.w), ÿиʽΪ(x,y,z)ر˳Ϊң
 */
bool configRefPlane(const Eigen::Vector3f& plane_point, const Eigen::Vector3f& plane_normal, const MatrixXf& camera_rays, MatrixXf& refPlane);

/*
 * for every screen pixel, compute their position in world coordinate system
 * params: Screen information, rotation of screen relative to world, translation of screen relative to world, matrix save screen pixels postion,
 * screen_pix_pos ʽ: 3 x (screen.h x screen.w), ÿиʽΪ(x,y,z)ر˳Ϊң
 */
bool configScreenPixelPos(const Screen& screen, const Matrix4f& WST, MatrixXf& screen_pix_pos);

/*
 * ĲƷͼ
 * ͼߴ磬xyڣ(size = 16˳ΪƵx,yƵx,y)
 */
// bool pattern_gen(std::vector<uint8_t> pat_size, std::vector<uint8_t> period, std::vector<MatrixXf>& pats);

/* four step phase shifting
 * params: images(size=16), result wraped phase map(size=4)
 */
bool phase_shifting(const std::vector<MatrixXf>& imgs, std::vector<MatrixXf>& wraped_ps_maps);

/*
 * two frequency phase unwrapping
 * params: λ(ƵƣƵƣƵƣƵ),ƵƵλ
 */
bool phase_unwrapping(const vector<MatrixXf>& wraped_ps_maps, uint16_t period_width, uint16_t period_height, vector<MatrixXf>& unwrap_ps_maps);

/*
 * ĻλimgӦصά
 * paramsλͼĻϢͼȷ߶ȷĻλã
 * imgλƥĻ꣨ʽ: 3 x (cam_img.w x cam_img.h), ÿиʽΪ(x,y,z)ر˳Ϊң£
 */
bool screen_camera_phase_match(const vector<MatrixXf>& unwrap_ps_maps, const Screen& screen, uint16_t period_width, uint16_t period_height,
    const MatrixXf& screen_pix_pos, MatrixXf& screen_camera_phase_match_pos);

/*
 * бʼ
 * params: ꣬οƽ棬Ļλƥ䣬бʽXY,ߴΪ1 x img.size
 */
bool slope_calculate(const Vector3f& camera_world, const MatrixXf& refPlane, const MatrixXf& screen_camera_phase_match_pos,
    std::vector<MatrixXfR>& slope);

/*
 * бʻָλmodal ʹ Zernike polynomials
 * params: xб(M x N)yбʣ߶Ƚxyߴ緶Χ(mm), Zernike polynomials 
 */
bool modal_reconstruction(const MatrixXf& sx, const MatrixXf& sy, MatrixXfR& Z, const std::vector<float>& range, uint32_t terms);

double computeReprojectionErrors(
    const vector<vector<cv::Point3f>>& objectPoints,
    const vector<vector<cv::Point2f>>& imagePoints,
    const vector<Mat>& rvecs, const vector<Mat>& tvecs,
    const Mat& cameraMatrix, const Mat& distCoeffs,
    vector<float>& perViewErrors);

/*
// aruco 궨
struct ArucoBoard {
    cv::Size markers_size;
    float markerLength;    // unit: m
    float markerSeparation;
    cv::Ptr<cv::aruco::Dictionary> dictionary;
    float screenOffset_X;
    float screenOffset_Y;
};


bool aruco_analyze(const vector<Mat>& imgs_board, const ArucoBoard& board, const Mat& cameraMatrix, const Mat& distCoeffs, vector<Mat>& Rmats,
    vector<cv::Vec3f>& Tvecs, vector<vector<int>>& Ids, vector<vector<vector<cv::Point2f>>>& corners,
    vector<vector<vector<cv::Point2f>>>& rejectedCandidates, std::string output_dir);
*/
/*
 * system geometry calibration
 * implement of paper "Flexible geometrical calibration for fringe-reflection
 measurement(2012)"
 */
bool system_calib(vector<Matrix3f>& CVR, vector<Vector3f>& CVTL, Matrix3f& CSR, VectorXf& CSTL_D, vector<Vector3f>& n);

class ArithPmd
{
public:
    ArithPmd();
    ~ArithPmd();

public:
    cv::Size img_size;      // ͼֱ
    uint16_t period_width;  // ȷ
    uint16_t period_height; // ߶ȷ

    cv::Mat mask0, mask1, mask2, mask3;
    
private:
    Screen screen;          // ĻϢ
    ROI roi0,roi1,roi2,roi3;                // ͼƬϵؽ

    Matrix4f CTC0,CTC1, CTC2, CTC3;

    MatrixXf refPlane0, refPlane1, refPlane2, refPlane3;
    Vector3f camera_origin_world0, camera_origin_world1, camera_origin_world2, camera_origin_world3;
    MatrixXf screen_pix_pos0, screen_pix_pos1, screen_pix_pos2, screen_pix_pos3;
    MatrixXf roi_m0, roi_m1, roi_m2, roi_m3; // mask ľ

    // std::vector<Matrix4f*> CTC{}; // תƽƾ

    // Ľ
    // std::vector<MatrixXf*> refPlane{};      // οƽĽ㣬ϵ£3 x (img.h x img.w)
    // std::vector <Vector3f*> camera_origin_world{ };   // ϵԭϵµıʾ
    // std::vector<MatrixXf*> screen_pix_pos{};// Ļϵµλã3 x (screen.h x screen.w)

    
    // std::vector<Mat> patterns;   // ͶӰͼ
    // std::vector<Eigen::MatrixXf> img_pats;  // ͼ
    // uint16_t screen_delay;  // ͶӰʱȴʱ()


public:

    int Initlization(const nlohmann::json& param);

    int Preprocess(const std::vector<cv::Mat>& images, int area_threshold, int32_t camera_index);

    int Predict(const std::vector<Eigen::MatrixXf>& img_pats, MatrixXfR& pcl, int32_t camera_index);

    int Stitch(const pcl::PointCloud<pcl::PointXYZ>& cloud1, const pcl::PointCloud<pcl::PointXYZ>& cloud2, pcl::PointCloud<pcl::PointXYZ>& cloud_output, int32_t camera_index);
};



#endif // ARITHPMD_H
